//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
//  Copyright (c) 2005-2007 Andre Merzky (andre@merzky.net)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef SAGA_SAGA_ADAPTORS_ADAPTOR_HPP
#define SAGA_SAGA_ADAPTORS_ADAPTOR_HPP

#include <boost/thread/thread.hpp>
#include <boost/thread/recursive_mutex.hpp>

#include <boost/plugin.hpp>
#include <boost/plugin/export_plugin.hpp>
#include <boost/thread.hpp>
#include <boost/tokenizer.hpp>
#include <boost/preprocessor/cat.hpp>
#include <boost/preprocessor/stringize.hpp>

#include <saga/adaptors/utils/utils.hpp>

#include <impl/engine/adaptor_selector.hpp>
#include <saga/uuid.hpp>

///////////////////////////////////////////////////////////////////////////////
#ifndef SAGA_ADAPTOR_NAME
# ifndef SAGA_ENGINE
#   error "SAGA_ADAPTOR_NAME needs to be defined for saga adaptor macros"
# else
#   define SAGA_ADAPTOR_NAME engine   // needed for get_name()
# endif
#endif

///////////////////////////////////////////////////////////////////////////////
#ifndef SAGA_ADAPTOR_LIB_NAME
#define SAGA_ADAPTOR_LIB_NAME                                                 \
        BOOST_PP_CAT(libsaga_adaptor_, SAGA_ADAPTOR_NAME )                    \
    /**/
#endif

///////////////////////////////////////////////////////////////////////////////
namespace saga 
{
    // definition of cpi names
    namespace adaptors
    {
        // general usage
        char const* const attribute_cpi        = "attribute_cpi";
        char const* const permissions_cpi      = "permissions_cpi";

        // file package
        char const* const namespace_entry_cpi  = "namespace_entry_cpi";
        char const* const namespace_dir_cpi    = "namespace_dir_cpi";
        char const* const file_cpi             = "file_cpi";
        char const* const directory_cpi        = "directory_cpi";
        
        // locical_file package
        char const* const logical_file_cpi      = "logical_file_cpi";
        char const* const logical_directory_cpi = "logical_directory_cpi";
        
        // job package
        char const* const job_cpi              = "job_cpi";
        char const* const service_cpi          = "service_cpi";
        
        // stream package
        char const* const stream_cpi           = "stream_cpi";
        char const* const server_cpi           = "server_cpi";

        // RPC package
        char const* const rpc_cpi              = "rpc_cpi";
        
        // information services package
        char const* const advert_cpi           = "advert_cpi";
        char const* const advert_directory_cpi = "advert_directory_cpi";
        
        // return the name of the implementing CPI for a saga object type
        inline std::string 
        get_cpi_name(saga::object::type type) 
        {
            switch (type) {
            case saga::object::NSEntry:           return namespace_entry_cpi;
            case saga::object::NSDirectory:       return namespace_dir_cpi;
            case saga::object::File:              return file_cpi;
            case saga::object::Directory:         return directory_cpi;
            case saga::object::LogicalFile:       return logical_file_cpi;
            case saga::object::LogicalDirectory:  return logical_directory_cpi;
            case saga::object::Job:               return job_cpi;
            case saga::object::JobService:        return service_cpi;
            case saga::object::Stream:            return stream_cpi;
            case saga::object::StreamService:     return server_cpi;
            case saga::object::RPC:               return rpc_cpi;
            case saga::object::Advert:            return advert_cpi;
            case saga::object::AdvertDirectory:   return advert_directory_cpi;
            default:
                break;
            }
            return "Unknown";
        }
    }
    
    /// @cond
    /** These methods are not within API scope */
    class adaptor
    {
    private:
        // this mutex is used to synchronize access to the adaptor member data
        typedef boost::recursive_mutex mutex_type;
        mutex_type mtx_;

    protected:
        // this uuid identifies the adaptor object
        saga::uuid adaptor_uuid_;
    
    public:
        virtual ~adaptor (void) {}
        virtual impl::adaptor_selector::adaptor_info_list_type 
                adaptor_register (saga::impl::session *) = 0;

        // support functions for the adaptor_data helper class
        void lock_data() 
        {
            boost::detail::thread::lock_ops<mutex_type>::lock(mtx_);
        }
        void unlock_data() 
        {
            boost::detail::thread::lock_ops<mutex_type>::unlock(mtx_);
        }
        
        // access the identifying uuid of this adaptor
        saga::uuid const & get_uuid (void) const
        {
            return adaptor_uuid_;
        }

        // access the identifying name of this adaptor
        virtual std::string get_name    (void) const = 0;
    };
    /// @endcond
  
///////////////////////////////////////////////////////////////////////////////
} // namespace saga

///////////////////////////////////////////////////////////////////////////////
//  Forward the SAGA_ADAPTOR_REGISTER macro to the correct BOOST_PLUGIN macro 
//  sequence
#define SAGA_ADAPTOR_REGISTER(ActualType)                                     \
        BOOST_PLUGIN_EXPORT_LIST (SAGA_ADAPTOR_LIB_NAME)                      \
        BOOST_PLUGIN_EXPORT      (SAGA_ADAPTOR_LIB_NAME, saga::adaptor,       \
                                  ActualType,        "adaptor");              \
    /**/

#endif // SAGA_SAGA_ADAPTORS_ADAPTOR_HPP

