//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef SAGA_ADAPTOR_TASK_V2_HPP
#define SAGA_ADAPTOR_TASK_V2_HPP

#include <saga/util.hpp>
#include <saga/task.hpp>
#include <impl/engine/task.hpp>
#include <impl/engine/wrapper_task.hpp>
#include <saga/uuid.hpp>

///////////////////////////////////////////////////////////////////////////////
namespace saga 
{ 
  namespace adaptors 
  {
    
    /**
      * Helper class to create saga::impl::task's.
      *
      * As this task is not a template class but has a template'd constructor
      * the templates are hidden from the adaptor programmer. Further this
      * class is not visible within the saga:: namespace, so templates
      * does not occur within the saga facade.
      */
    template <typename RetVal>
    class task : public saga::task<RetVal>
    {
    private:
        typedef saga::task<RetVal> base_type;
        
        /**
          * template'd helper function, which creates a appropriate
          * saga::impl::task object
          */
        template<typename Cpi, typename FuncBase, typename FuncRetV>
        static inline saga::impl::task<Cpi, FuncBase, FuncRetV, RetVal> *
        create_task(char const *func_name, 
                    TR1::shared_ptr<Cpi> cpi_, 
                    void (FuncBase::*f)(FuncRetV),
                    RetVal func_ret,
                    bool (FuncBase::*prep_func)(FuncRetV, saga::uuid))
        {
            return new saga::impl::task<Cpi, FuncBase, FuncRetV, RetVal>(
                func_name, cpi_, f, func_ret, prep_func);
        }
        
        
        /**
          * template'd helper function, which creates a appropriate
          * saga::impl::task object
          *
          * The function to execute takes as a further parameter,
          * more precisely, a uuid of the task, which will be executed in the adaptor
          *
          */
        template<typename Cpi, typename FuncBase, typename FuncRetV>
        static inline saga::impl::wrapper_task<Cpi, FuncBase, FuncRetV, RetVal> *
        create_task(char const                 *func_name, 
                    TR1::shared_ptr<Cpi>      cpi_, 
                    void (FuncBase::*f)        (FuncRetV, saga::uuid),
                    RetVal                      func_ret,
                    bool (FuncBase::*prep_func)(FuncRetV, saga::uuid))
        {
            return new saga::impl::wrapper_task<Cpi, FuncBase, FuncRetV, RetVal>(
                func_name, cpi_, f, func_ret, prep_func);
        }

    public:
        task(void);
        
        ~task (void) 
        {
        }
        
        /**
          * template'd constructor
          *
          * calls member function create_task and puts the returned task
          * into the saga::task class, which is part of the facade.
          */ 
        template<typename Cpi, typename FuncBase, typename FuncRetV>
        task (char const                  * func_name,
              TR1::shared_ptr <Cpi>       cpi_, 
              void (FuncBase::*func)      ( FuncRetV ),
              RetVal                        func_ret,
              bool (FuncBase::*prep_func) ( FuncRetV, saga::uuid ) = NULL) 
            : base_type (create_task (func_name, cpi_, func, 
                                      func_ret, prep_func))
        {
        }
        
        /**
          * overloaded template'd constructor
          *
          * calls member function create_task and puts the returned task
          * into the saga::task class, which is part of the facade.
          * Differs from the other constructor, because the func takes a 
          * further parameter, a task uuid.
          *
          */ 
        template<typename Cpi, typename FuncBase, typename FuncRetV>
        task (char const                  * func_name,
              TR1::shared_ptr <Cpi>       cpi_, 
              void (FuncBase::*func)      ( FuncRetV, saga::uuid ),
              RetVal                        func_ret,
              bool (FuncBase::*prep_func) ( FuncRetV, saga::uuid ) = NULL) 
            : base_type (TR1::shared_ptr <saga::impl::task_base>
                           (create_task (func_name, cpi_, func, 
                                         func_ret, prep_func))) 
        {
        }

        ///////////////////////////////////////////////////////////////////////
        // bring in the generated constructors with 1 to N parameters.
        #include <saga/adaptors/task_impl_v2.hpp>
    };

  } // adaptors

} // namespace saga
///////////////////////////////////////////////////////////////////////////////

#endif // SAGA_ADAPTOR_TASK_V2_HPP

