//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef SAGA_SAGA_DETAIL_PERMISSIONS_HPP
#define SAGA_SAGA_DETAIL_PERMISSIONS_HPP

#include <string>

// include dependent spec sections
#include <saga/util.hpp>
#include <saga/base.hpp>
#include <saga/monitorable.hpp>
#include <saga/call.hpp>

// suppress warnings about dependent classes not being exported from the dll
#if defined(BOOST_MSVC)
#pragma warning(push)
#pragma warning(disable : 4251 4231 4660)
#endif

///////////////////////////////////////////////////////////////////////////////
// for modules not belonging to the engine attributes should not be exported
#if !defined (SAGA_CREATE_PREPROCESSED_FILES) && defined (BOOST_WINDOWS) 
#  ifdef     SAGA_ENGINE_EXPORTS
#    define  SAGA_EXPORT_PERMISSIONS   __declspec (dllexport) 
#  elif !defined (SAGA_NO_EXPORT_PERMISSIONS)
#    define  SAGA_EXPORT_PERMISSIONS   __declspec (dllimport) 
#  endif
#endif

#if !defined (SAGA_EXPORT_PERMISSIONS)
#    define  SAGA_EXPORT_PERMISSIONS   /* empty */
#endif

///////////////////////////////////////////////////////////////////////////////
namespace saga { namespace detail
    {
        /// @cond
        ///////////////////////////////////////////////////////////////////////////
        template <typename Derived, typename Tag>
        struct SAGA_EXPORT_PERMISSIONS permissions_priv
        {
            static saga::task_type permissions_allow (Derived const& this_, 
                                                      saga::void_t&, std::string id, int p);
            static saga::task_type permissions_deny (Derived const& this_, 
                                                     saga::void_t&, std::string id, int p);
            static saga::task_type permissions_check (Derived const& this_, 
                                                      bool&, std::string id, int p);
            
            static saga::task_type get_owner (Derived const& this_, std::string& r);
            static saga::task_type get_group (Derived const& this_, std::string& r);
        };
        /// @endcond
        ///////////////////////////////////////////////////////////////////////////
        
        
        /*! \brief   The %permissions interface.
         *
         *  \details Detailed %description starts here.
         */
        template <typename Derived>
        struct SAGA_EXPORT_PERMISSIONS permissions
        {
        private:
            template <typename Tag>
            saga::task_type permissions_allowpriv (saga::void_t& r, std::string id, int p, Tag) const
            { return permissions_priv<Derived, Tag>::permissions_allow(derived(), r, id, p); }
            
            template <typename Tag>
            saga::task_type permissions_denypriv (saga::void_t& r, std::string id, int p, Tag) const
            { return permissions_priv<Derived, Tag>::permissions_deny(derived(), r, id, p); }
            
            template <typename Tag>
            saga::task_type permissions_checkpriv (bool& r, std::string id, int p, Tag) const
            { return permissions_priv<Derived, Tag>::permissions_check(derived(), r, id, p); }
            
            template <typename Tag>
            saga::task_type get_ownerpriv (std::string& r, Tag) const
            { return permissions_priv<Derived, Tag>::get_owner(derived(), r); }
            
            template <typename Tag>
            saga::task_type get_grouppriv (std::string& r, Tag) const
            { return permissions_priv<Derived, Tag>::get_group(derived(), r); }
            
        public:
            ///@cond
            impl::permissions_interface* get_perm (void);
            impl::permissions_interface* get_perm (void) const;
            ///@endcond
            
        protected:
            ///@cond
            Derived& derived() 
            { return static_cast<Derived&>(*this); }
            Derived const& derived() const 
            { return static_cast<Derived const&>(*this); }
            ///@endcond
        public: 
            
            /*! \brief     Set a permission flag for this %object instance.
             *  \param     permission_id The permission to set the flag for
             *  \param     flag The flag to set
             *  \exception saga::exception (on condition)
             *
             *  \details   Detailed %description starts here.
             */       
            saga::void_t permissions_allow(std::string permission_id, int flag) {
                saga::void_t ret;
                permissions_allowpriv (ret, permission_id, flag, saga::task_base::Sync()); 
                return ret;
            }
            ///@cond
            SAGA_CALL_PUB_2_DEF_0(saga::void_t, permissions_allow, std::string, int);
            ///@endcond
            
            /*! \brief     Remove a permission flag from this %object instance.
             *  \param     permission_id The permission to remove the flag from
             *  \param     flag The flag to remove
             *  \exception saga::exception (on condition)
             *
             *  \details   Detailed %description starts here.
             */       
            saga::void_t permissions_deny(std::string permission_id, int flag) {
                saga::void_t ret;
                permissions_denypriv (ret, permission_id, flag, saga::task_base::Sync()); 
                return ret;
            }   
            ///@cond
            SAGA_CALL_PUB_2_DEF_0(saga::void_t, permissions_deny, std::string, int);
            ///@endcond
            
            /*! \brief     Check this %object instance for a permission flag.
             *  \param     permission_id The permission to check for a flag 
             *  \param     flag The flag to check for
             *  \exception saga::exception (on condition)
             *
             *  \details   Detailed %description starts here.
             */           
            bool permissions_check(std::string permission_id, int flag) const {
                bool ret;
                permissions_checkpriv (ret, permission_id, flag, saga::task_base::Sync()); 
                return ret;
            }   
            ///@cond
            SAGA_CALL_CONST_PUB_2_DEF_0(bool, permissions_check, std::string, int);
            ///@endcond
            
            /*! \brief     Returns the owner of this %object instance.
             *  \return    The owner's name
             *  \exception saga::exception (on condition)
             *
             *  \details   Detailed %description starts here.
             */       
            std::string get_owner() const {
                std::string ret;
                get_ownerpriv (ret, saga::task_base::Sync()); 
                return ret;
            }   
            ///@cond
            SAGA_CALL_CONST_PUB_0_DEF_0(std::string, get_owner);
            ///@endcond
            
            /*! \brief     Returns the group of this %object instance.
             *  \return    The group's name
             *  \exception saga::exception (on condition)
             *
             *  \details   Detailed %description starts here.
             */       
            std::string get_group() const {
                std::string ret;
                get_grouppriv (ret, saga::task_base::Sync()); 
                return ret;
            }   
            ///@cond        
            SAGA_CALL_CONST_PUB_0_DEF_0(std::string, get_group);
            ///@endcond
        };
        
        ///////////////////////////////////////////////////////////////////////////////
    }}

#endif


