//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef SAGA_IMPL_ENGINE_REGISTER_MEMBERS_HPP
#define SAGA_IMPL_ENGINE_REGISTER_MEMBERS_HPP

#include <boost/mpl/if.hpp>

#include <saga/adaptors/adaptor.hpp>
#include <impl/engine/cpi_info.hpp>
#include <saga/adaptors/adaptor.hpp>

///////////////////////////////////////////////////////////////////////////////
namespace saga { namespace impl {

///////////////////////////////////////////////////////////////////////////////
//
//  The set of base_from_func meta template functions below is used to 
//  return the type of the class a certain given function pointer type is 
//  implemented in.
//
//  The different overloads are required for different numbers of parameters 
//  these functions take.
//
//  Note: using the variadics template features (as proposed for C++0x) this 
//        could be rewritten as:
//
//          template <typename T>
//          struct base_from_func;
//
//          template <typename RT, typename Base, typename... Args>
//          struct base_from_func<RT (Base::*)(Args...)>
//          {
//              typedef Base type;
//          };
//
template <typename T>
struct base_from_func;

template <typename RT, typename Base>
struct base_from_func<RT (Base::*)()>
{
    typedef Base type;
};

template <typename RT, typename Base, typename T1>
struct base_from_func<RT (Base::*)(T1)>
{
    typedef Base type;
};

template <typename RT, typename Base, typename T1, typename T2>
struct base_from_func<RT (Base::*)(T1, T2)>
{
    typedef Base type;
};

template <typename RT, typename Base, typename T1, typename T2, typename T3>
struct base_from_func<RT (Base::*)(T1, T2, T3)>
{
    typedef Base type;
};

template <
    typename RT, typename Base, typename T1, typename T2, typename T3, 
    typename T4
>
struct base_from_func<RT (Base::*)(T1, T2, T3, T4)>
{
    typedef Base type;
};

template <
    typename RT, typename Base, typename T1, typename T2, typename T3, 
    typename T4, typename T5
>
struct base_from_func<RT (Base::*)(T1, T2, T3, T4, T5)>
{
    typedef Base type;
};

template <
    typename RT, typename Base, typename T1, typename T2, typename T3, 
    typename T4, typename T5, typename T6
>
struct base_from_func<RT (Base::*)(T1, T2, T3, T4, T5, T6)>
{
    typedef Base type;
};

///////////////////////////////////////////////////////////////////////////////
//
//  The register_member_helper and dont_register_member_helper types are pure
//  helpers allowing to dispatch the registration of a certain member function
//  either to no-ops or to the real registration.
//
template <typename Cpi>
struct register_member_helper
{
    // register synchronous CPI function
    static bool 
    eval_sync(saga::impl::v1_0::cpi_info& info, char const* name, 
        saga::impl::v1_0::preference_type const& prefs)
    {
        if (Cpi::register_sync_member(name)) 
            info.add_op(saga::impl::v1_0::op_info (name, prefs));
        return true;
    }
    
    // register asynchronous CPI function
    static bool 
    eval_async(saga::impl::v1_0::cpi_info& info, char const* name, 
        saga::impl::v1_0::preference_type const& prefs)
    {
        if (Cpi::register_async_member(name))
            info.add_op(saga::impl::v1_0::op_info (name, prefs));
        return true;
    }
};

struct dont_register_member_helper
{
    static bool 
    eval_sync(saga::impl::v1_0::cpi_info&, char const*, 
        saga::impl::v1_0::preference_type const&) 
    {
        return false;   // don't do anything
    }
    
    static bool 
    eval_async(saga::impl::v1_0::cpi_info&, char const*, 
        saga::impl::v1_0::preference_type const&) 
    {
        return false;   // don't do anything
    }
};

///////////////////////////////////////////////////////////////////////////////
//
//  The functions register_member_sync() and register_member_async() use some
//  meta template programming tricks to figure out, whether the given member 
//  function 'Func' is implemented in a derived class of the given 'BaseCpi' 
//  type.
//
//  If the function _is_ implemented, then it will be registered with the SAGA
//  engine. If it is implemented in the corresponding base CPI only (see
//  base_from_func meta template function), the function will not be registered 
//  with the SAGA engine).
//
//  The function returns true if the function is defined in a derived class, 
//  otherwise it returns false.
//
template <typename BaseCpi, typename Func>
inline bool 
register_member_sync(Func f, saga::impl::v1_0::cpi_info& info, 
    char const* name, saga::impl::v1_0::preference_type const& prefs)
{
    typedef typename base_from_func<Func>::type func_base_type;
    typedef typename boost::mpl::if_<
        TR1::is_base_of<BaseCpi, func_base_type>,
            register_member_helper<func_base_type>, 
            dont_register_member_helper
    >::type register_type;
    
    return register_type::eval_sync(info, name, prefs);
}

template <typename BaseCpi, typename Func>
inline bool 
register_member_async(Func f, saga::impl::v1_0::cpi_info& info, 
    char const* name, saga::impl::v1_0::preference_type const& prefs)
{
    typedef typename base_from_func<Func>::type func_base_type;
    typedef typename boost::mpl::if_<
        TR1::is_base_of<BaseCpi, func_base_type>,
            register_member_helper<func_base_type>, 
            dont_register_member_helper
    >::type register_type;
    
    return register_type::eval_async(info, name, prefs);
}

///////////////////////////////////////////////////////////////////////////////
}}  // saga::impl

///////////////////////////////////////////////////////////////////////////////
//  Register the given CPI function for the given CPI implementation.
#define SAGA_REGISTER_MEMBER(retval, info, base_cpi, cpi, func, prefs)        \
        retval = saga::impl::register_member_sync<base_cpi>(                  \
            &cpi::sync_ ## func, info, "sync_" #func, prefs) || retval;       \
        retval = saga::impl::register_member_async<base_cpi>(                 \
            &cpi::async_ ## func, info, "async_" #func, prefs) || retval;     \
    /**/

#define SAGA_REGISTER_MEMBER_EX(retval, info, base_cpi, funcname, prefs,      \
                                syncfunc, asyncfunc)                          \
        retval = saga::impl::register_member_sync<base_cpi>(                  \
            syncfunc, info, "sync_" #funcname, prefs) || retval;              \
        retval = saga::impl::register_member_async<base_cpi>(                 \
            asyncfunc, info, "async_" #funcname, prefs) || retval;            \
    /**/

/* FIXME: add bulk functions to all cpi's                                  \ */
/*     if (cpi::register_bulk_members())                                   \ */
/*         saga::impl::register_member<cpi>(&cpi::bulk_ ## func, info,     \ */
/*             "bulk_" #func, prefs);                                      \ */

#endif // !SAGA_IMPL_ENGINE_REGISTER_MEMBERS_HPP

