#ifndef SAGA_PACKAGES_JOB_JOB_HPP
#define SAGA_PACKAGES_JOB_JOB_HPP

#if defined(__WAVE__) && defined(SAGA_CREATE_PREPROCESSED_FILES)
#pragma wave option(preserve: 2, line: 1, output: "preprocessed/job.hpp")
#endif
//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
//  Copyright (c) 2007 Ole Weidner (oweidner@cct.lsu.edu)
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#if defined(__WAVE__) && defined(SAGA_CREATE_PREPROCESSED_FILES)
#pragma wave option(output: null)
#endif

#include <string>
#include <vector>
#include <iosfwd>

// include dependent spec sections
#include <saga/util.hpp>
#include <saga/base.hpp>
#include <saga/types.hpp>
#include <saga/session.hpp>
#include <saga/call.hpp>

#include <saga/packages/job/config.hpp>
#include <saga/packages/job/job_description.hpp>
#include <saga/packages/job/istream.hpp>
#include <saga/packages/job/ostream.hpp>

#include <saga/detail/attribute.hpp>
#include <saga/detail/permissions.hpp>

// suppress warnings about dependent classes not being exported from the dll
#if defined(BOOST_MSVC)
#pragma warning(push)
#pragma warning(disable: 4251 4231 4275 4660)
#endif

#ifdef SAGA_DEBUG
#include <saga/packages/job/preprocessed/job.hpp>
#else

#if defined(__WAVE__) && defined(SAGA_CREATE_PREPROCESSED_FILES)
#pragma wave option(preserve: 2, line: 1, output: "preprocessed/job.hpp")
#endif
///////////////////////////////////////////////////////////////////////////////
namespace saga
{
    /*! \brief The <I>Job Management</I> API package
     *
     *   Detailed %description here...
     */   
    namespace job {
        
    /*! \brief  Available %attributes defined for saga::job::job and 
     *         saga::job::description instances.
     *
     *   <b>NOTE:</b> Attributes with a <i>description_</i>prefix are read/write 
     *   %attributes and belong to a saga::job::description instance. The rest are 
     *   read-only %attributes and available for saga::job::job instances.<br><br>
     *   Although JSDL based %attribute names are used for the %job description
     *   %attributes, the API uses no explicit representation of JSDL (i.e. JSDL
     *   compliant XML).<br>
     *   <br><b>Usage example:</b><br><br>
     *   <code>saga::job::service js("localhost");<br><br>
     *   saga::job::description jd();<br>
     *   jd.set_attribute("description_executable", "/bin/uname");<br><br>
     *   saga::job::job j1 = js.create_job(jd);<br>
     *   std::cout << j1.get_attribute("executionhosts") << std::endl;</code>
     *   
     */     
    namespace attributes 
    {
        //  read only attributes
        /*! \brief <b>[r]</b> SAGA representation of the %job identifier
         *         (format: [backend %url]-[native id])
         */
        char const* const jobid = "JobID";
        /*! \brief <b>[r]</b> List of host names or IP addresses allocated to run this %job 
         */
        char const* const executionhosts = "ExecutionHosts";
        /*! \brief <b>[r]</b> Time stamp of the %job creation in the resource manager 
         */
        char const* const created = "Created";
        /*! \brief <b>[r]</b> Time stamp indicating when the %job started running 
         */
        char const* const started = "Started";
        /*! \brief <b>[r]</b> Time stamp indicating when the %job completed 
         */
        char const* const finished = "Finished";
        /*! \brief <b>[r]</b> Working directory on the execution host 
         */
        char const* const workingdirectory = "WorkingDirectory";
        /*! \brief <b>[r]</b> Process exit code as collected by the wait(2) series of 
         *         system calls
         */
        char const* const exitcode = "ExitCode";
        /*! \brief <b>[r]</b> Signal number which caused the %job to exit <br><br>
         */
        char const* const termsig = "Termsig";
    }

    /*! \brief   Available %metrics defined for saga::job::job instances.
     *
     *   Several metrics on the saga::job::job class (the class implements the 
     *   saga::monitorable interface) reflect %attributes from the %job 
     *   description. This redundancy is intentional, and aims at providing 
     *   information about (a) %attributes which may change at runtime, and (b) 
     *   %attributes for jobs for which no %job description is available (e.g.
     *   saga::job::job instances obtained by calling saga::job::service::get_job().<br>
     *   <br><b>Usage example:</b><br><br>
     *   <code>//TODO</code>
     */    
    namespace metrics 
    { 
        /*! \brief Fires on state changes of the job, and has the literal 
         *         value of the job state enum
         */
        char const* const statedetail = "job.StateDetail";
        /*! \brief Fires as a job receives a signal, and has a value 
         *         indicating the signal number
         */
        char const* const signal = "job.Signal";
        /*! \brief Number of CPU seconds consumed by the job
         */
        char const* const cputimelimit = "job.CPUTimeLimit";
        /*! \brief Current aggregate memory usage
         */
        char const* const memoryuse = "job.MemoryUse";
        /*! \brief Current aggregate virtual memory usage
         */
        char const* const vmemoryuse = "job.VmemoryUse";
        /*! \brief Current performance (FLOPS)
         */
        char const* const performance = "job.Performance";
    }

    /*! \brief Brief %description starts here     
     *
     *   Detailed %description starts here.
     */
    enum state 
    {
        Unknown   =  saga::task_base::Unknown,   // -1
        New       =  saga::task_base::New,       //  0
        Running   =  saga::task_base::Running,   //  1
        Failed    =  saga::task_base::Failed,    //  2
        Done      =  saga::task_base::Done,      //  3
        Canceled  =  saga::task_base::Canceled,  //  4
        Suspended =  5
    };    
    
    /*! \brief Brief %description starts here     
     *
     *   Detailed %description starts here.
     */
    class SAGA_JOB_PACKAGE_EXPORT job 
        : public saga::task_type,
          public saga::detail::attribute<job>,
          public saga::detail::permissions<job>
    {

    private:
          
        friend struct detail::attribute<job>;  // needs to access get_impl()
        friend struct detail::permissions<job>;

        // job inspection
        SAGA_CALL_PRIV_0 (std::string, get_job_id)
        SAGA_CALL_PRIV_0 (saga::job::state, get_state)
        SAGA_CALL_PRIV_0 (description, get_description)

        // The following functions are a bit special because they do not provide
        // the usual plain sync interface function returning the result. We have
        // the task based variants available only.
        SAGA_CALL_PRIV_0 (saga::job::ostream, get_stdin)
        SAGA_CALL_PRIV_0 (saga::job::istream, get_stdout)
        SAGA_CALL_PRIV_0 (saga::job::istream, get_stderr)

        // job management
        SAGA_CALL_PRIV_0 (saga::void_t, suspend)
        SAGA_CALL_PRIV_0 (saga::void_t, resume)
        SAGA_CALL_PRIV_0 (saga::void_t, checkpoint)
        SAGA_CALL_PRIV_1 (saga::void_t, migrate, description)
        SAGA_CALL_PRIV_1 (saga::void_t, signal, int)

    protected:
        /// @cond
        /** These methods are not within API scope */
        typedef detail::attribute<job> attribute_base;
        TR1::shared_ptr <saga::impl::job> get_impl (void) const;
        explicit job (saga::impl::job *);
        /// @endcond
          
    public:
        /*! \brief Creates the object.     
         *
         *   Detailed %description starts here.
         */
        job();

        /*! \brief Creates the object.     
         *
         *   Detailed %description starts here.
         */
        explicit job (saga::object const& o);

        /*! \brief Destroys the object.     
         *
         *   Detailed %description starts here.
         */
        ~job();

        /*! \brief Brief %description starts here     
         *
         *   Detailed %description starts here.
         */
        job &operator= (saga::object const& o);

        /*! \brief Brief %description starts here     
         *  \exception saga::exception (on condition)
         *
         *   Detailed %description starts here.
         */
        std::string get_job_id(void) {
            std::string ret;
            get_job_idpriv (ret, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_0_DEF_0 (std::string, get_job_id)

        /*! \brief Get the state of the job.  
         *  \return State of the job
         *  \exception saga::exception (on condition)
         *
         *   Detailed %description starts here.
         */
        saga::job::state get_state(void) {
            saga::job::state ret;
            get_statepriv (ret, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_0_DEF_0 (saga::job::state, get_state)

        /*! \brief Retrieve the description which was used to submit this 
         *         job instance. 
         *  \exception saga::exception (on condition)
         *  \return A job_description object
         *
         *   Detailed %description starts here.
         */
        saga::job::description get_description(void) {
            saga::job::description ret;
            get_descriptionpriv (ret, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_0_DEF_0 (description, get_description)

        /*! \brief Retrieve input stream for a job     
         *  \exception saga::exception (on condition)
         *  \return standard input stream for the job 
         *
         *   Detailed %description starts here.
         */
        saga::job::ostream get_stdin(void) {
            saga::job::ostream ret;
            get_stdinpriv (ret, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_0_DEF_0 (saga::job::ostream, get_stdin)

        /*! \brief Retrieve output stream of job    
         *  \exception saga::exception (on condition)
         *  \return Standard output stream for the job 
         *
         *   Detailed %description starts here.
         */
        saga::job::istream get_stdout(void) {
            saga::job::istream ret;
            get_stdoutpriv (ret, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_0_DEF_0 (saga::job::istream, get_stdout)

        /*! \brief Retrieve error stream of job   
         *  \exception saga::exception (on condition)
         *  \return Standard error stream for the job 
         *
         *   Detailed %description starts here.
         */
        saga::job::istream get_stderr(void) {
            saga::job::istream ret;
            get_stderrpriv (ret, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_0_DEF_0 (saga::job::istream, get_stderr)

        /*! \brief Ask the resource manager to suspend this job   
         *  \exception saga::exception (on condition)
         *
         *   Detailed %description starts here.
         */
        saga::void_t suspend(void) {
            saga::void_t ret;
            suspendpriv (ret, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_0_DEF_0 (saga::void_t, suspend)

        /*! \brief Ask the resource manager to resume this job   
         *  \exception saga::exception (on condition)
         *
         *   Detailed %description starts here.
         */
        saga::void_t resume(void) {
            saga::void_t ret;
            resumepriv (ret, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_0_DEF_0 (saga::void_t, resume)

        /*! \brief Ask the resource manager to checkpoint this job    
         *  \exception saga::exception (on condition)
         *
         *   Detailed %description starts here.
         */
        saga::void_t checkpoint(void) {
            saga::void_t ret;
            checkpointpriv (ret, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_0_DEF_0 (saga::void_t, checkpoint)

        /*! \brief Ask the resource manager to migrate this job    
         *  \exception saga::exception (on condition)
         *  \param job_desc New job parameters to apply when the job is migrated 
         *
         *   Detailed %description starts here.
         */
        saga::void_t migrate(description job_desc) {
            saga::void_t ret;
            migratepriv (ret, job_desc, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_1_DEF_0 (saga::void_t, migrate, description)

        /*! \brief Ask the resource manager to deliver an arbitrary signal to 
         *         this job
         *  \exception saga::exception (on condition)
         *  \param signum Signal number to be delivered 
         *
         *   Detailed %description starts here.
         */
        saga::void_t signal(int signal) {
            saga::void_t ret;
            signalpriv (ret, signal, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_1_DEF_0 (saga::void_t, signal, int)
    };

    }   // namespace job

}   // namespace saga

#if defined(__WAVE__) && defined(SAGA_CREATE_PREPROCESSED_FILES)
#pragma wave option(output: null)
#endif

#endif // !defined(SAGA_DEBUG)

// re-enable warnings about dependent classes not being exported from the dll
#if defined(BOOST_MSVC)
#pragma warning(pop)
#endif

#endif // !defined(SAGA_PACKAGES_JOB_JOB_HPP)

