//  Copyright (c) 2005-2007 Andre Merzky (andre@merzky.net)
//  Copyright (c) 2005-2007 Hartmut Kaiser (hartmut.kaiser@gmail.com)
//  Copyright (c) 2007 Ole Weidner (oweidner@cct.lsu.edu)
// 
//  Use, modification and distribution is subject to the Boost Software
//  License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt)

#ifndef SAGA_COMM_STREAM_STREAM_HPP
#define SAGA_COMM_STREAM_STREAM_HPP

// include stl
#include <string>
#include <vector>

// include dependent spec sections
#include <saga/util.hpp>
#include <saga/call.hpp>
#include <saga/base.hpp>
#include <saga/session.hpp>
#include <saga/task.hpp>
#include <saga/buffer.hpp>
#include <saga/detail/attribute.hpp>
#include <saga/detail/monitorable.hpp>

// include package config
#include <saga/packages/stream/config.hpp>

// suppress warnings about dependent classes not being exported from the dll
#if defined(BOOST_MSVC)
#pragma warning(push)
#pragma warning(disable: 4251 4231 4660)
#endif

namespace saga
{
    /*! \brief The <I>Streams</I> API package
     */
    namespace stream {
    
    /*! \brief   Available %attributes defined for saga::stream instances.
     */
    namespace attributes 
    {
        //  read write attributes
        char const* const stream_bufsize = "Bufsize";
        char const* const stream_timeout = "Timeout";
        char const* const stream_blocking = "Blocking";
        char const* const stream_compression = "Compression";
        char const* const stream_nodelay = "Nodelay";
        char const* const stream_reliable = "Reliable";
    }
    
    /*! \brief   Available metrices defined for saga::stream instances.
     */
    namespace metrics 
    {
        char const* const stream_state = "stream.State";
        char const* const stream_read = "stream.Read";
        char const* const stream_write = "stream.Write";
        char const* const stream_exception = "stream.Exception";
        char const* const stream_dropped = "stream.Dropped";
    }

    /*! \brief The <I>Streams</I> API package 
     *
     *   Detailed %description here...
     */  

  
    /*! \brief Brief %description starts here     
     *
     *   Detailed %description starts here.
     */
    enum state
    {
        Unknown      = -1,
        New          =  1,
        Open         =  2,
        Closed       =  3,
        Dropped      =  4,
        Error        =  5
    };

    /*! \brief Brief %description starts here     
     *
     *   Detailed %description starts here.
     */     
    enum activity
    {
        Read         =  1,
        Write        =  2,
        Exception    =  4
    };

    /*! \brief Brief %description starts here     
     *
     *   Detailed %description starts here.
     */
    class SAGA_STREAM_PACKAGE_EXPORT stream 
        : public saga::object,
          public saga::detail::attribute<stream>,
          public saga::detail::monitorable<stream>
    {
        /// @cond
        /** These methods are not within API scope */
        friend struct saga::detail::attribute<stream>;    // needs to access get_impl()
        friend struct saga::detail::monitorable<stream>;  // needs to access get_impl()

        typedef saga::detail::attribute<stream> attribute_base_type;
        typedef saga::detail::monitorable<stream> monitorable_base_type;
        /// @endcond
      
    public:

      /*! \brief Brief %description starts here     
       *
       *   Detailed %description starts here.
       */
      //typedef stream_base::state state; 

    protected:
        /// @cond
        /** These methods are not within API scope */
        TR1::shared_ptr <saga::impl::stream> get_impl (void) const;
        explicit stream (saga::impl::stream *);
        /// @endcond

    private:
        // stream inspection methods
        SAGA_CALL_CONST_PRIV_0 (std::string,   get_url     )
        SAGA_CALL_CONST_PRIV_0 (saga::context, get_context )

        // stream methods
        SAGA_CALL_PRIV_0 (saga::context, connect)
        SAGA_CALL_PRIV_2 (std::vector<saga::stream::activity>, wait, saga::stream::activity, double)
        SAGA_CALL_PRIV_1 (saga::void_t, close, double)
        SAGA_CALL_PRIV_2 (saga::ssize_t, read, saga::mutable_buffer, saga::ssize_t)
        SAGA_CALL_PRIV_2 (saga::ssize_t, write, saga::const_buffer, saga::ssize_t)

        void init_attributes();
        void init_metrics();
      
    public:
        /*! \brief Initializes a client client stream, for later connection to an server.      
         *
         *   Detailed %description starts here.
         */
        stream (session const & s, std::string const & url = "");

        /*! \brief Initializes a client client stream, for later connection to an server.       
         *
         *   Detailed %description starts here.
         */
        explicit stream (std::string const & url);

        /*! \brief Initializes a client client stream, for later connection to an server.        
         *
         *   Detailed %description starts here.
         */
        stream ();

        /*! \brief Initializes a client client stream, for later connection to an server.        
         *
         *   Detailed %description starts here.
         */
        stream (saga::object const& o);

        /*! \brief Destroys an stream object     
         *
         *   Detailed %description starts here.
         */
        ~stream (void);

        /*! \brief Brief %description starts here     
         *
         *   Detailed %description starts here.
         */
        stream &operator= (saga::object const& o);

        /*! \brief Get the URL used for creating this stream     
         *  \exception saga::exception (on condition)
         *  \return String containing the URL of the connection. 
         *
         *   Detailed %description starts here.
         */
        std::string get_url(void) {
            std::string ret;
            get_urlpriv (ret, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_CONST_PUB_0_DEF_0 (std::string, get_url)

        /*! \brief Return remote authorization info     
         *  \exception saga::exception (on condition)
         *  \return Remote context
         *
         *   Detailed %description starts here.
         */
        saga::context get_context(void) {
            saga::context ret;
            get_contextpriv (ret, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_CONST_PUB_0_DEF_0 (saga::context, get_context )

        /*! \brief Establishes a connection to the target defined during the 
         *         construction of the stream   
         *  \exception saga::exception (on condition)
         *
         *   Detailed %description starts here.
         */
        saga::context connect(void) {
            saga::context ret;
            connectpriv (ret, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_0_DEF_0 (saga::context, connect)

        /*! \brief Check if stream is ready for reading/writing, or if it has 
         *         entered an error state  
         *  \exception saga::exception (on condition)
         *  \param what Parameter list of activity types to wait for 
         *  \param timeout Number of seconds to wait 
         *  \return activity type causing the call to return 
         *
         *   Detailed %description starts here.
         */
        std::vector<saga::stream::activity> wait(saga::stream::activity what,
                                                 double timeout=-1.0) 
        {
            std::vector<saga::stream::activity> ret;
            waitpriv (ret, what, timeout, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_2_DEF_1 (std::vector<saga::stream::activity>, wait, saga::stream::activity, double, -1.0)

        /*! \brief Closes an active connection    
         *  \exception saga::exception (on condition)
         *
         *   Detailed %description starts here.
         */     
        saga::void_t close(double timeout=0.0) {
            saga::void_t ret;
            closepriv (ret, timeout, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_1_DEF_1 (saga::void_t, close, double, 0.0)

        /*! \brief Read a raw buffer from stream    
         *  \exception saga::exception (on condition)
         *  \param length Maximum number of bytes that can be copied in to the buffer. 
         *  \param buffer Empty buffer passed in to get filled 
         *  \return Number of bytes read, if successful. (0 is also valid)
         *
         *   Detailed %description starts here.
         */
        saga::ssize_t read(saga::mutable_buffer buffer, saga::ssize_t length) {
            saga::ssize_t ret;
            readpriv (ret, buffer, length, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_2_DEF_0 (saga::ssize_t, read, saga::mutable_buffer, saga::ssize_t)

        /*! \brief Write a raw buffer to stream    
         *  \exception saga::exception (on condition)
         *  \param length Number of bytes of data in the buffer 
         *  \param buffer Raw array containing data that will be sent out via socket 
         *  \return Bytes written if successful 
         *
         *   Detailed %description starts here.
         */
        saga::ssize_t write(saga::const_buffer buffer, saga::ssize_t length) {
            saga::ssize_t ret;
            writepriv (ret, buffer, length, saga::task_base::Sync()); 
            return ret;
        }
        SAGA_CALL_PUB_2_DEF_0 (saga::ssize_t, write, saga::const_buffer, saga::ssize_t)
    };

    }
} 

// re-enable warnings about dependent classes not being exported from the dll
#if defined(BOOST_MSVC)
#pragma warning(pop)
#endif

#endif // SAGA_COMM_STREAM_STREAM_HPP

