
#include "shell.hpp"


///////////////////////////////////////////////////////////
//
// we get some environment variables at startup, for the
// initial working directory and initial job manager.
//
std::string init_contact (void)
{
    char const* env = ::getenv ("SAGA_SHELL_CONTACT");

    return env ? env : "fork://localhost/";
}

std::string init_host (std::string url)
{
    saga::url u (url);
    std::string host = u.get_host ();

    if ( "" == host )
    {
      return ("localhost");
    }

    return (host);
}

std::string init_home (void)
{
    char const* env = ::getenv ("SAGA_SHELL_HOME");

    return env ? env : "file://localhost//tmp";
}


///////////////////////////////////////////////////////////
//
// the c'tor only initializes a number of variables
//
shell::shell (void)
  : session_ (),
    home_    (init_home ()),
    cwd_     (session_, home_),
    contact_ (init_contact ()),
    host_    (init_host (contact_)),
    js_      (session_, contact_)
    
{
  return;
}


///////////////////////////////////////////////////////////
//
// now and then, we have to re-initialize member variables, 
// as some operations change the environment
//
void shell::init (void)
{
  home_    = init_home    ();
  contact_ = init_contact ();
  host_    = init_host    (contact_);

  carp ("reinit js with " + contact_);

  // recreate session
  saga::session s;
  session_ = s;

  // recreate job service in that session
  saga::job::service js (s, contact_);
  js_ = js;

  // recreate cwd in that session
  saga::filesystem::directory cwd (s, cwd_.get_url ());
  cwd_ = cwd;
}


///////////////////////////////////////////////////////////
//
// as long as there is input, get it, and run the 
// respective command
//
void shell::run (void)
{
  // get first command line
  prompt ();
  std::string input;
  
  getline (std::cin, input);

  // keep running until we see 'exit'
  while ( input != "exit" )
  {
    // split line in command and arguments
    std::string command = input;
    std::string args    = input;

    try 
    {

      size_t idx = input.find (" ");

      if ( idx != 0  && idx != std::string::npos)
      {
        command.erase (idx);
        args.erase    (0, idx + 1);
      }
      else
      {
        args = "";
      }

      // operations on the name space
      if      ( command == "ls"      ) { c_ls      (args); }
      else if ( command == "pwd"     ) { c_pwd     (args); }
      else if ( command == "cd"      ) { c_cd      (args); }
      else if ( command == "cp"      ) { c_cp      (args); }
      else if ( command == "mv"      ) { c_mv      (args); }
      else if ( command == "ln"      ) { c_ln      (args); }
      else if ( command == "rm"      ) { c_rm      (args); }
      else if ( command == "mkdir"   ) { c_mkdir   (args); }
      else if ( command == "rmdir"   ) { c_rmdir   (args); }

      // operations on file contents 
      else if ( command == "cat"     ) { c_cat     (args); }

      // job related operations
      else if ( command == "run"     ) { c_run     (args); }
      else if ( command == "ps"      ) { c_ps      (args); }
      else if ( command == "status"  ) { c_status  (args); }
      else if ( command == "suspend" ) { c_suspend (args); }
      else if ( command == "resume"  ) { c_resume  (args); }
      else if ( command == "kill"    ) { c_kill    (args); }

      // others
      else if ( command == "set"     ) { c_set     (args); }
      else                             { c_help    (args); }
    }

    // catch SAGA exceptions, but leave system exceptions untouched
    catch ( saga::exception const & e ) 
    {
      switch ( e.get_error () )
      {
        // handle NotImplemented gracefully
        case saga::NotImplemented:
          carp ("This method is not yet implemented in saga-a");
          break;

        // generic handler and dump of other exceptions
        default:
          carp ("saga exception caught: " + std::string (e.what ()));
          break;
      }
    }

    // get ready for next command
    prompt ();
    getline (std::cin, input);
  }

  return;
}

///////////////////////////////////////////////////////////
//
// print a shell prompt, with optional message
//
void shell::prompt (std::string text, 
                    std::string marker)
{
  if ( text != "" )
  {
    std::cout << text << std::endl;
  }

  std::cout << marker << std::flush;
}


///////////////////////////////////////////////////////////
//
// print an error message
//
void shell::carp (std::string text)
{
  std::cout << "Ooops!" << std::endl;

  if ( text != "" )
  {
    std::cout << text << std::endl;
  }
}


///////////////////////////////////////////////////////////
//
// translate a job state into a string
//
std::string shell::state_to_string (saga::job::state state)
{
  switch ( state )
  {
    case saga::job::New       : return ("New      ");
    case saga::job::Running   : return ("Running  ");
    case saga::job::Suspended : return ("Suspended");
    case saga::job::Done      : return ("Done     ");
    case saga::job::Failed    : return ("Failed   ");
    case saga::job::Canceled  : return ("Canceled ");
    default                   : 
        std::cout << "Unknown state: " << state << std::endl;
        return ("Unknown  ");
  }
}

